==> MarkXCC.Java-11.3.2-src.zip: com/marklogic/xcc/impl/handlers/PDCloudRequestController.java <==
/*
 * Copyright (c) 2006-2025 Progress Software Corporation and/or its subsidiaries or affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.marklogic.xcc.impl.handlers;

import com.marklogic.http.HttpChannel;
import com.marklogic.xcc.Request;
import com.marklogic.xcc.RequestOptions;
import com.marklogic.xcc.ResultSequence;
import com.marklogic.xcc.exceptions.PDCloudRequestException;
import com.marklogic.xcc.exceptions.RequestException;
import com.marklogic.xcc.impl.ContentSourceImpl;
import com.marklogic.xcc.impl.Credentials;
import com.marklogic.xcc.impl.SessionImpl;
import com.marklogic.xcc.spi.ConnectionProvider;
import com.marklogic.xcc.spi.ServerConnection;

import java.io.IOException;
import java.net.HttpURLConnection;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

public class PDCloudRequestController extends AbstractRequestController {
    private static final int MAX_RETRY = 5;
    private static final int RETRY_DELAY_MILLIS = 125;
    private static final Map<Integer, ResponseHandler> handlers =
        new HashMap<>();
    private Credentials.PDCloudAuthConfig pdCloudAuthConfig;
    private String boundary;

    static {
        addDefaultHandler(handlers, new UnrecognizedCodeHandler());
        // 200 OK --> Request success
        addHandler(handlers, HttpURLConnection.HTTP_OK,
            new PDCloudSuccessResponseHandler());
        // 401 Unauthorized --> Invalid API key specified/session token expired
        addHandler(handlers, HttpURLConnection.HTTP_UNAUTHORIZED,
            new UnauthorizedHandler());
        // 403 Access Denied --> Access denied to the given resource
        addHandler(handlers, HttpURLConnection.HTTP_FORBIDDEN,
            new PDCloudForbiddenHandler());
        // 404 Not Found --> Invalid request URL
        addHandler(handlers, HttpURLConnection.HTTP_NOT_FOUND,
            new NotFoundCodeHandler());
        // 410 HTTP Gone --> Tenant expiry
        addHandler(handlers, HttpURLConnection.HTTP_GONE,
            new PDCloudGoneHandler());
        //5xx General server errors
        addHandler(handlers, HttpURLConnection.HTTP_INTERNAL_ERROR,
            new ServerExceptionHandler());
        addHandler(handlers, HttpURLConnection.HTTP_BAD_GATEWAY,
            new ServiceUnavailableHandler());
        addHandler(handlers, HttpURLConnection.HTTP_UNAVAILABLE,
            new ServiceUnavailableHandler());
        addHandler(handlers, HttpURLConnection.HTTP_GATEWAY_TIMEOUT,
            new ServiceUnavailableHandler());
    }

    public PDCloudRequestController(
        Credentials.PDCloudAuthConfig pdCloudAuthConfig) {
        super(handlers, DEFAULT_SERVER_PATH,
            pdCloudAuthConfig.getTokenEndpoint());
        this.pdCloudAuthConfig = pdCloudAuthConfig;
    }

    @Override
    public ResultSequence runRequest(
        ConnectionProvider provider, Request request, Logger logger)
        throws RequestException {
        SessionImpl session = (SessionImpl)request.getSession();
        RequestOptions options = request.getEffectiveOptions();
        ServerConnection connection = null;
        RequestException re = null;
        int t = 0;
        for (; t < MAX_RETRY; t++) {
            if (t > 0) {
                if (logger.isLoggable(Level.FINE)) {
                    logger.log(Level.FINE, "Retrying connecting to Progress " +
                        "Data Cloud (" + t + ").");
                }
            }
            try {
                sleepFor(interTryDelay(RETRY_DELAY_MILLIS, t), logger);
                connection = provider.obtainConnection(session, request, logger);
                ResultSequence rs = serverDialog(connection, request, options,
                    logger);
                if ((rs == null) || rs.isCached()) {
                    provider.returnConnection(connection, logger);
                }
                return rs;
            } catch (RequestException e) {
                logger.log(Level.WARNING, "Request exception connecting to " +
                    "Progress Data Cloud. Cannot obtain session token.", e);
                provider.returnConnection(connection, logger);
                if (e instanceof PDCloudRequestException) {
                    if (!e.isRetryable()) throw e;
                    if (logger.isLoggable(Level.FINE)) {
                        logger.log(Level.FINE, "Retryable exception caught.", e);
                    }
                    re = e;
                } else throw e;
            } catch (IOException e) {
                logger.log(Level.WARNING, "Connection IOException caught. " +
                    "Cannot obtain session token.", e);
                if (connection != null) {
                    provider.returnErrorConnection(connection, e, logger);
                }
                re = new PDCloudRequestException(e.getMessage(), request, e,
                    true);
            } catch (Exception e) {
                logger.log(Level.WARNING, "Exception connecting to " +
                    "Progress Data Cloud. Cannot obtain session token.", e);
                re = new PDCloudRequestException(e.getMessage(), request, e,
                    true);
            }
        }
        logger.log(Level.WARNING, "Automatic connecting to Progress Data Cloud " +
            "retries (" + t + ") exhausted, throwing: " + re, re);
        throw re;
    }

    @Override
    public ResultSequence serverDialog(ServerConnection connection,
                                       Request request, RequestOptions options,
                                       Logger logger)
        throws IOException, RequestException{
        ResultSequence rs = null;
        try{
            SessionImpl session = (SessionImpl)request.getSession();
            byte[] bodyBytes = buildFormBody().getBytes();
            String reqUri = addTokenDurationToPath(httpPath);
            HttpChannel http = buildChannel(connection, reqUri, session,
                bodyBytes.length, options, logger);
            http.write(bodyBytes);
            if (HttpChannel.isUseHTTP()) {
                http.write("\r\n".getBytes());
            }
            int code = http.getResponseCode();
            ResponseHandler handler = findHandler(code);
            rs = (ResultSequence)handler.handleResponse(http, code, request, pdCloudAuthConfig, logger);
        } finally {
            if(rs!=null)
                rs.close();
        }
        return null;
    }

    private HttpChannel buildChannel(ServerConnection connection, String path,
                                     SessionImpl session, int bufferSize,
                                     RequestOptions options, Logger logger) {
        String method = "POST";
        HttpChannel http = new HttpChannel(connection.channel(), method, path,
            bufferSize, options.getTimeoutMillis(), logger);

        // Add MLCloud headers
        ContentSourceImpl contentSource =
            (ContentSourceImpl)session.getContentSource();
        if(HttpChannel.isUseHTTP()) {
            ConnectionProvider cp = contentSource.getConnectionProvider();
            http.setRequestHeader("Host", cp.getHostName() + ":" + cp.getPort());
        }
        http.setRequestHeader("User-Agent", session.userAgentString());
        http.setRequestHeader("Accept", session.getAcceptedContentTypes());
        http.setRequestContentType("multipart/form-data; boundary=" + boundary);
        return http;
    }

    private String buildFormBody() {
        HttpChannel.MultipartFormBody formBody =
            new HttpChannel.MultipartFormBody();
        this.boundary = formBody.getBoundary();
        formBody.addTextBody("key",
            new String(pdCloudAuthConfig.getApiKey()));
        formBody.addTextBody("grant_type", pdCloudAuthConfig.getGrantType());
        return formBody.buildFormBody();
    }

    private String addTokenDurationToPath(String httpPath) {
        StringBuilder sb = new StringBuilder(httpPath);
        int tokenDuration = pdCloudAuthConfig.getTokenDuration();
        if (tokenDuration != Credentials.DEFAULT_TOKEN_DURATION) {
            sb.append("?duration=");
            sb.append(tokenDuration);
        }
        return sb.toString();
    }
}

==> MarkXCC.Java-12.0.1-src.zip: com/marklogic/xcc/impl/handlers/PDCloudRequestController.java <==
/*
 * Copyright (c) 2006-2025 Progress Software Corporation and/or its subsidiaries or affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.marklogic.xcc.impl.handlers;

import com.marklogic.http.HttpChannel;
import com.marklogic.xcc.Request;
import com.marklogic.xcc.RequestOptions;
import com.marklogic.xcc.ResultSequence;
import com.marklogic.xcc.exceptions.PDCloudRequestException;
import com.marklogic.xcc.exceptions.RequestException;
import com.marklogic.xcc.impl.ContentSourceImpl;
import com.marklogic.xcc.impl.Credentials;
import com.marklogic.xcc.impl.SessionImpl;
import com.marklogic.xcc.spi.ConnectionProvider;
import com.marklogic.xcc.spi.ServerConnection;

import java.io.IOException;
import java.net.HttpURLConnection;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

public class PDCloudRequestController extends AbstractRequestController {
    private static final int MAX_RETRY = 5;
    private static final int RETRY_DELAY_MILLIS = 125;
    private static final Map<Integer, ResponseHandler> handlers =
        new HashMap<>();
    private Credentials.PDCloudAuthConfig pdCloudAuthConfig;
    private String boundary;

    static {
        addDefaultHandler(handlers, new UnrecognizedCodeHandler());
        // 200 OK --> Request success
        addHandler(handlers, HttpURLConnection.HTTP_OK,
            new PDCloudSuccessResponseHandler());
        // 401 Unauthorized --> Invalid API key specified/session token expired
        addHandler(handlers, HttpURLConnection.HTTP_UNAUTHORIZED,
            new UnauthorizedHandler());
        // 403 Access Denied --> Access denied to the given resource
        addHandler(handlers, HttpURLConnection.HTTP_FORBIDDEN,
            new PDCloudForbiddenHandler());
        // 404 Not Found --> Invalid request URL
        addHandler(handlers, HttpURLConnection.HTTP_NOT_FOUND,
            new NotFoundCodeHandler());
        // 410 HTTP Gone --> Tenant expiry
        addHandler(handlers, HttpURLConnection.HTTP_GONE,
            new PDCloudGoneHandler());
        //5xx General server errors
        addHandler(handlers, HttpURLConnection.HTTP_INTERNAL_ERROR,
            new ServerExceptionHandler());
        addHandler(handlers, HttpURLConnection.HTTP_BAD_GATEWAY,
            new ServiceUnavailableHandler());
        addHandler(handlers, HttpURLConnection.HTTP_UNAVAILABLE,
            new ServiceUnavailableHandler());
        addHandler(handlers, HttpURLConnection.HTTP_GATEWAY_TIMEOUT,
            new ServiceUnavailableHandler());
    }

    public PDCloudRequestController(
        Credentials.PDCloudAuthConfig pdCloudAuthConfig) {
        super(handlers, DEFAULT_SERVER_PATH,
            pdCloudAuthConfig.getTokenEndpoint());
        this.pdCloudAuthConfig = pdCloudAuthConfig;
    }

    @Override
    public ResultSequence runRequest(
        ConnectionProvider provider, Request request, Logger logger)
        throws RequestException {
        SessionImpl session = (SessionImpl)request.getSession();
        RequestOptions options = request.getEffectiveOptions();
        ServerConnection connection = null;
        RequestException re = null;
        int t = 0;
        for (; t < MAX_RETRY; t++) {
            if (t > 0) {
                if (logger.isLoggable(Level.FINE)) {
                    logger.log(Level.FINE, "Retrying connecting to Progress " +
                        "Data Cloud (" + t + ").");
                }
            }
            try {
                sleepFor(interTryDelay(RETRY_DELAY_MILLIS, t), logger);
                connection = provider.obtainConnection(session, request, logger);
                ResultSequence rs = serverDialog(connection, request, options,
                    logger);
                if ((rs == null) || rs.isCached()) {
                    provider.returnConnection(connection, logger);
                }
                return rs;
            } catch (RequestException e) {
                logger.log(Level.WARNING, "Request exception connecting to " +
                    "Progress Data Cloud. Cannot obtain session token.", e);
                provider.returnConnection(connection, logger);
                if (e instanceof PDCloudRequestException) {
                    if (!e.isRetryable()) throw e;
                    if (logger.isLoggable(Level.FINE)) {
                        logger.log(Level.FINE, "Retryable exception caught.", e);
                    }
                    re = e;
                } else throw e;
            } catch (IOException e) {
                logger.log(Level.WARNING, "Connection IOException caught. " +
                    "Cannot obtain session token.", e);
                if (connection != null) {
                    provider.returnErrorConnection(connection, e, logger);
                }
                re = new PDCloudRequestException(e.getMessage(), request, e,
                    true);
            } catch (Exception e) {
                logger.log(Level.WARNING, "Exception connecting to " +
                    "Progress Data Cloud. Cannot obtain session token.", e);
                re = new PDCloudRequestException(e.getMessage(), request, e,
                    true);
            }
        }
        logger.log(Level.WARNING, "Automatic connecting to Progress Data Cloud " +
            "retries (" + t + ") exhausted, throwing: " + re, re);
        throw re;
    }

    @Override
    public ResultSequence serverDialog(ServerConnection connection,
                                       Request request, RequestOptions options,
                                       Logger logger)
        throws IOException, RequestException{
        SessionImpl session = (SessionImpl)request.getSession();
        byte[] bodyBytes = buildFormBody().getBytes();
        String reqUri = addTokenDurationToPath(httpPath);
        HttpChannel http = buildChannel(connection, reqUri, session,
            bodyBytes.length, options, logger);
        http.write(bodyBytes);
        if (HttpChannel.isUseHTTP()) {
            http.write("\r\n".getBytes());
        }
        int code = http.getResponseCode();
        ResponseHandler handler = findHandler(code);
        handler.handleResponse(http, code, request, pdCloudAuthConfig, logger);
        return null;
    }

    private HttpChannel buildChannel(ServerConnection connection, String path,
                                     SessionImpl session, int bufferSize,
                                     RequestOptions options, Logger logger) {
        String method = "POST";
        HttpChannel http = new HttpChannel(connection.channel(), method, path,
            bufferSize, options.getTimeoutMillis(), logger);

        // Add MLCloud headers
        ContentSourceImpl contentSource =
            (ContentSourceImpl)session.getContentSource();
        if(HttpChannel.isUseHTTP()) {
            ConnectionProvider cp = contentSource.getConnectionProvider();
            http.setRequestHeader("Host", cp.getHostName() + ":" + cp.getPort());
        }
        http.setRequestHeader("User-Agent", session.userAgentString());
        http.setRequestHeader("Accept", session.getAcceptedContentTypes());
        http.setRequestContentType("multipart/form-data; boundary=" + boundary);
        return http;
    }

    private String buildFormBody() {
        HttpChannel.MultipartFormBody formBody =
            new HttpChannel.MultipartFormBody();
        this.boundary = formBody.getBoundary();
        formBody.addTextBody("key",
            new String(pdCloudAuthConfig.getApiKey()));
        formBody.addTextBody("grant_type", pdCloudAuthConfig.getGrantType());
        return formBody.buildFormBody();
    }

    private String addTokenDurationToPath(String httpPath) {
        StringBuilder sb = new StringBuilder(httpPath);
        int tokenDuration = pdCloudAuthConfig.getTokenDuration();
        if (tokenDuration != Credentials.DEFAULT_TOKEN_DURATION) {
            sb.append("?duration=");
            sb.append(tokenDuration);
        }
        return sb.toString();
    }
}
